//------------------------------------------------------------------------------
//                      Copyright Tech-Know-Ware Ltd, 2002
//------------------------------------------------------------------------------


// File name:   umf-lib.c
//------------------------------------------------------------------------------
// Description: This file contains library functions for reading UMF encoded
//				messages.
//
// Notice:      WARNING: This is an early release of this code, and as such may
//				contain a number of bugs.  If you find a bug, please report it
//				to:
//
//					umf-lib-bugs@tech-know-ware.com
//
//				This file is presented 'as is' and comes with no warrenty of 
//				any kind.  
//
//              This file, or any derivative of it, may be used for any 
//				purpose, including commercial purposes.
//
//              Users are permitted to modify the contents of the file, as 
//				long as this notice and the copyright notice remains.
//
//              Users are permitted to extract elements of this file into
//              other files, subject to this notice and the copyright notice 
//              also being copied into the target files.
//
//              Tech-Know-Ware Ltd shall not be liable for any damages suffered 
//              by anyone or anything due to the use of this file however they 
//				may be caused.
//
//				For more information on UMF, go to:
//
//					http://www.tech-know-ware.com/umf
//
//              Copyright Tech-Know-Ware Ltd, 2001, 2002.
//
//              End of Notice.
//
// Module Tag:  umf - Unified Message Format
// Version:     V.95.6
// Entry Points: 
//				fumf_is_message_complete
//				fumf_parse_message
//				fumf_set_location
//				fumf_restore_location
//				fumf_is_present
//				fumf_read_void
//				fumf_read_boolean
//				fumf_read_integer
//				fumf_read_ipv4
//				fumf_read_ipv6 - Not in this release
//				fumf_read_date
//				fumf_read_time
//				fumf_read_ascii
//				fumf_read_ascii_dynamic
//				fumf_read_const
//				fumf_read_unicode
//				fumf_read_unicode_dynamic
//				fumf_read_byte_array
//				fumf_read_byte_array_dynamic
//				fumf_read_embedded
//				fumf_read_embedded_dynamic
//				fumf_read_union
//				fumf_read_enumerated_union
//				fumf_release
//
// Changes:
// P. Cordell	2001-06-04	Preparation for evaluation
// P. Cordell	2001-06-17	V0.95.3 - Changed error code convention so that 
//							functions that fail return 0.
// P. Cordell	2002-03-07	V.95.4 Release
// P. Cordell	2002-04-07	Move to V.95.5
// P. Cordell	2002-04-07	Changed fumf_read_null to fumf_read_void
// P. Cordell	2002-04-25	Move to V.95.6
//
// To-Do:		


#include <stdlib.h>
#include <ctype.h>
#include <string.h>


#include "umf-lib.h"

#define KUMF_EOM			(0)
#define KUMF_MAX_RECURSIONS	100	/* This may need to be adjusted according to the message definition */

#define MUMF_ASSERT(x)		/* Users of this code may define this */


#define MUMF_IS_FIRST_TOKEN_CHAR( x ) ( \
								(x) > ' ' &&  /* No WS, no Ctrl */	\
								! strchr( ",\"'{}()^-", (x) ) &&	\
								(x) != KUMF_EOM )

#define MUMF_IS_FIRST_TAG_CHAR( x ) ( \
								(x) > ' ' &&  /* No WS, no Ctrl */	\
								! strchr( ",\"'{}()^-", (x) ) &&	\
								! isdigit( (x) ) &&                 \
								(x) != KUMF_EOM )

#define MUMF_IS_TOKEN_CHAR( x, include_equals ) ( \
								(x) > ' ' &&  /* No WS, no Ctrl */	\
								! strchr( "}),", (x) ) &&			\
								((x) != '=' || include_equals) &&	\
								(x) != KUMF_EOM )


// These are for the position data file
const char gumf_is_struct[] = "s";
const char gumf_is_union[] = "u";
const char gumf_is_untagged[] = "";


static int p_parse( sumf_parse_ctrl *ap_parse_ctrl, sumf_message_element **app_new_element );
static void p_free_element_info( sumf_message_element *ap_element );
tumf_bool fumf_initialise( sumf_parse_ctrl *ap_parse_ctrl );


//-----------------------------------------------------------------------------
//          The following functions handle reading the message input.
//     They are designed so that it is easy to alter whether the message is
//                      stored in memory or in a file.
//-----------------------------------------------------------------------------


static tumf_input_location p_get_input_location( sumf_parse_ctrl *ap_parse_ctrl )
{
    return ap_parse_ctrl->p_next - ap_parse_ctrl->p_message;
}


static void p_set_input_location( sumf_parse_ctrl *ap_parse_ctrl,
									tumf_input_location a_location )
{
    ap_parse_ctrl->p_next = ap_parse_ctrl->p_message + a_location;
	ap_parse_ctrl->unget_char = '\0';
}


// Function name:   p_unget_char
//-----------------------------------------------------------------------------
// Description: Ungetting a char is handled by storing the ungot char rather
//				than by adjusting the ap_parse_ctrl->p_next pointer because
//				the char being returned could be a space substituted
//				for a comment.  In this case, updating ap_parse_ctrl->p_next
//				does not work.

static void p_unget_char( sumf_parse_ctrl *ap_parse_ctrl, int a_c )
{
	ap_parse_ctrl->unget_char = a_c;
}


// Function name:   p_get_char
//-----------------------------------------------------------------------------
// Description: Reads next characeter from file.  This function can be called 
//				directly if comment removal is not required.

static int p_get_char( sumf_parse_ctrl *ap_parse_ctrl )
{
	if( ap_parse_ctrl->unget_char != '\0' )
		{
		int l_return_char = ap_parse_ctrl->unget_char;

		ap_parse_ctrl->unget_char = '\0';

		return l_return_char;
		}

	if( ap_parse_ctrl->p_next < ap_parse_ctrl->p_message_end )
		return *(ap_parse_ctrl->p_next++);

	return KUMF_EOM;
}


static int p_get_unicode_char( sumf_parse_ctrl *ap_parse_ctrl )
{
	int l_char;
	int l_c = p_get_char( ap_parse_ctrl );

	if( l_c <= 0x7f )
		l_char = l_c & 0xff;

	else if( l_c <= 0xc0 )
		{
		// 2 bytes
		l_char = (l_c & 0x1f) << 6;
		l_c = p_get_char( ap_parse_ctrl );
		l_char |= l_c & 0x3f;
		}

	else if( l_c <= 0xe0 )
		{
		// 3 bytes
		l_char = (l_c & 0x1f) << 6;
		l_c = p_get_char( ap_parse_ctrl );
		l_char |= l_c & 0x3f;
		l_char <<= 6;
		l_c = p_get_char( ap_parse_ctrl );
		l_char |= l_c & 0x3f;
		}

	else
		return KUMF_EOM;		// Can not do UCS-4, so stop parsing

	if( l_c == KUMF_EOM )
		return KUMF_EOM;

	return l_char;
}


static int p_skip_comment_to_nl( sumf_parse_ctrl *ap_parse_ctrl )
{
    int l_c;

    do  
        {
        l_c =  p_get_char( ap_parse_ctrl );   // Read next charater

        } while( l_c != '\n' && l_c != KUMF_EOM );    
                                                // Until comment end

    l_c = ' ';     // Treat a comment as single white space

    return l_c;
}


static int p_skip_comment( sumf_parse_ctrl *ap_parse_ctrl )
{
    int l_c;
    int l_extra = '\0';     // Initialise to some non-valid character

    do  
        {
        l_c = l_extra;								// Store used character
        l_extra = p_get_char( ap_parse_ctrl );		// Read next charater
        if( l_extra == KUMF_EOM )                   // Check not at end of file
            {
            break;
            }
        } while( l_c != '*' || l_extra != '/' );    
                                                    // Until comment end

    l_c = ' ';     // Treat a comment as single white space


    return l_c;
}


// Function name: p_read
//-----------------------------------------------------------------------------
// Description: Reads input and removes any comments before passing a 
//              character that is not part of a comment back to the calling
//              function.  
//              Call p_get_char directly if it is desired not to remove things
//              that look like comments (This may occur in strings).

static int p_read( sumf_parse_ctrl *ap_parse_ctrl )
{
    int     l_c;

    if( (l_c = p_get_char( ap_parse_ctrl )) == KUMF_EOM )
        return KUMF_EOM;


    // Check for comments - Put in while loop to allow back-to-back comments
    while( l_c == '/' )
        {
        int     l_extra;

        l_extra = p_get_char( ap_parse_ctrl );				// Read next character from file
        if( l_extra == '/' )								// If this is a comment then
            l_c = p_skip_comment_to_nl( ap_parse_ctrl );	// -skip the comment
		else if( l_extra == '*' )
			l_c = p_skip_comment( ap_parse_ctrl );
        else
            {
            p_unget_char( ap_parse_ctrl, l_extra );

            break;
            }
        }

    return l_c;                  // Character nothing to do with comment so
                                    // -return it
}


//-----------------------------------------------------------------------------
//      Start of functions for parsing message into linked list tree 
//                         used to locate parameters.
//-----------------------------------------------------------------------------

static sumf_message_element *p_new_message_element()
{ 
	sumf_message_element *lp_new_element = 
			(sumf_message_element *)malloc( sizeof( sumf_message_element ));

    lp_new_element->p_tag = NULL;
    lp_new_element->tag_location = -1;
    lp_new_element->value_location = -1;
    lp_new_element->p_compound = NULL;
    lp_new_element->p_next = NULL;

	return lp_new_element;
}


static int p_read_token( sumf_parse_ctrl *ap_parse_ctrl, int a_c,
							char *ac_token, int a_token_size )
{
	int l_i;
	int l_include_equals = 0;

	MUMF_ASSERT( a_token_size > 0 );

	a_token_size--;		// Make space for termination

	if( a_c == '^' ) l_include_equals = 1;

    for( l_i=0; 
			l_i < a_token_size && MUMF_IS_TOKEN_CHAR( a_c, l_include_equals );
			l_i++ )
        {
		ac_token[l_i] = a_c;

		a_c = p_read( ap_parse_ctrl );
        }


	ac_token[l_i] = '\0';


	// Skip anything that won't fit into array
	while( MUMF_IS_TOKEN_CHAR( a_c, l_include_equals ) )
		a_c = p_read( ap_parse_ctrl );

	return a_c;
}


static void p_skip_value( sumf_parse_ctrl *ap_parse_ctrl, int a_c )
{
    if( a_c == '"' || a_c == '\'' || a_c == '(' )
        {
        int l_end_char = a_c;
        int l_escaped = FALSE;

		if( a_c == '(' ) l_end_char = ')';		// Set embedded end value

        while( (a_c = (l_end_char != '"') ? 
							p_get_char( ap_parse_ctrl ) : 
							p_get_unicode_char( ap_parse_ctrl )) != KUMF_EOM && 
					(a_c != l_end_char || l_escaped) )
            {
            if( !l_escaped && a_c == '\\' )
                l_escaped = TRUE;

            else if( l_escaped )
                l_escaped = FALSE;
            }
        }

	else
		{
		int l_include_equals = 0;

		if( a_c == '^' ) l_include_equals = 1;

        while( (a_c = p_read( ap_parse_ctrl )) != KUMF_EOM && 
				MUMF_IS_TOKEN_CHAR( a_c, l_include_equals ) )
            {
            }

		p_unget_char( ap_parse_ctrl, a_c );
		}
}


// Func name:   p_parse_value
//------------------------------------------------------------------------------
// Notes:		Locate-Correct - When we locate a value that is also a tag, we
//				have actually encountered a nested union.  It is important that
//				we set the tag_number to 0 so that the parameter location works
//				properly for this case.  See p_locate_parameter description.

static int p_parse_value( sumf_parse_ctrl *ap_parse_ctrl, 
							sumf_message_element *ap_element )
{
	int l_c;

	if( ap_parse_ctrl->recursion++ > KUMF_MAX_RECURSIONS )
		{
		ap_parse_ctrl->error = EUMF_TOO_MANY_RECURSIONS;
		return FALSE;
		}

	l_c = p_read( ap_parse_ctrl );		// Move past '='

	while( isspace( l_c ) )
		l_c = p_read( ap_parse_ctrl );

	ap_element->value_location = p_get_input_location( ap_parse_ctrl ) - 1;

	// Handle value appropriately
    if( l_c == KUMF_EOM || l_c < ' ' || l_c == '}' || l_c == '=' || l_c ==',' )
		{
		ap_parse_ctrl->error = EUMF_PARSE_FAILED;

		return FALSE;
		}

	else if( l_c == '"' || l_c == '\'' || l_c == '(' || l_c == '^' ||
					l_c == '-' || isdigit( l_c ) )
		{
		p_skip_value( ap_parse_ctrl, l_c );
		}

	else if( l_c == '{' )
		{
		if( ! p_parse( ap_parse_ctrl, &(ap_element->p_compound) ) )
			return FALSE;
		}

	else if( !isspace( l_c ) )
		{
		char lc_value[KUMF_MAX_TAG];
		tumf_input_location l_value_location = p_get_input_location( ap_parse_ctrl ) - 1;

		l_c = p_read_token( ap_parse_ctrl, l_c, lc_value, sizeof( lc_value ) );

		// Now look for possible following '='
		while( isspace( l_c ) )
			l_c = p_read( ap_parse_ctrl );

		if( l_c == '=' )
			{
			// This is something like tag = value&tag = value
			if( ! (ap_element->p_compound = p_new_message_element()) )
				{
				ap_parse_ctrl->allocation_failed = TRUE;

				return FALSE;
				}

			if( ! (ap_element->p_compound->p_tag = strdup( lc_value ) ) )
				{
				ap_parse_ctrl->allocation_failed = TRUE;

				return FALSE;
				}

			ap_element->p_compound->tag_location = l_value_location;
			ap_element->p_compound->tag_number = 0;			// See Locate-Correct

			if( ! p_parse_value( ap_parse_ctrl, ap_element->p_compound ) )
				return FALSE;
			}

		else
			p_unget_char( ap_parse_ctrl, l_c );
		}

	ap_parse_ctrl->recursion--;

	return TRUE;
}


// Func name:   p_parse
//------------------------------------------------------------------------------
// Notes:		Tag-modify - When we assign a tag number to a parameter, we
//				assume that the next parameter will not have the same tag 
//				number and increment l_tag_number accordingly.  If we 
//				subsequently encounter a comma (indicating that the next 
//				parameter should have the same tag number as the previous 
//				parameter, then we correct l_tag_number by decrementing it.
//				Note also that we have to assign the tag number after we've
//				tested for the comma so that the corrected value is used.
//				Hence the assignment is done at the end of the loop rather
//				than with all the other lp_new_element initialisation code.

static int p_parse( sumf_parse_ctrl *ap_parse_ctrl, 
					sumf_message_element **app_new_seg_head )
{
    int l_c;
    sumf_message_element *lp_segment_tail = NULL;
    char *lp_last_tag = "";
	tumf_input_location l_last_tag_location = -1;
	int l_tag_number = 0;
	int l_first_time = TRUE;


	if( ap_parse_ctrl->recursion++ > KUMF_MAX_RECURSIONS )
		{
		ap_parse_ctrl->error = EUMF_TOO_MANY_RECURSIONS;
		return FALSE;
		}

    while( 1 )
        {
        char lc_tag[KUMF_MAX_TAG];
        sumf_message_element *lp_new_element;

        lc_tag[0] = '\0';

        while( (l_c = p_read( ap_parse_ctrl )) != KUMF_EOM && isspace( l_c ) )
            {
            }

        if( l_c == '}' )
            break;

		else if( l_c == KUMF_EOM )
			{
			if( ap_parse_ctrl->recursion == 1 )
				break;		// Case where transport indicates end of message

			else
				{
				ap_parse_ctrl->error = EUMF_MESSAGE_TOO_SHORT;

				return FALSE;
				}
			}

		else if( l_c < ' ' || l_c == '=' || l_first_time == TRUE && l_c ==',' )
			{
			ap_parse_ctrl->error = EUMF_PARSE_FAILED;

			return FALSE;
			}

		l_first_time = FALSE;

        if( ! (lp_new_element = p_new_message_element()) )
            {
            ap_parse_ctrl->allocation_failed = TRUE;

            return FALSE;
            }

        if( lp_segment_tail )
            lp_segment_tail->p_next = lp_new_element;
        else
            *app_new_seg_head = lp_new_element;

        lp_segment_tail = lp_new_element;


		if( l_c == '"' || l_c == '\'' || l_c == '(' || 		// Value with no tag
					l_c == '^' || l_c == '-' || isdigit( l_c ) )
			{
			lp_new_element->value_location = p_get_input_location( ap_parse_ctrl ) - 1;
			lp_new_element->tag_number = l_tag_number;

			p_skip_value( ap_parse_ctrl, l_c );
			}

		else if( l_c == '{' )						// Value with no tag
			{
			lp_new_element->value_location = p_get_input_location( ap_parse_ctrl ) - 1;
			if( ! p_parse( ap_parse_ctrl, &(lp_new_element->p_compound) ) )
				return FALSE;
			}

		else if( l_c == ',' )
			{
			if( ! lp_last_tag )
				lp_new_element->p_tag = NULL;

			else if( ! (lp_new_element->p_tag = strdup( lp_last_tag ) ) )
				{
				ap_parse_ctrl->allocation_failed = TRUE;

				return FALSE;
				}

			lp_new_element->tag_location = l_last_tag_location;

			if( l_tag_number > 0 ) l_tag_number--;		// See Tag-modify

			if( ! p_parse_value( ap_parse_ctrl, lp_new_element ) )
				return FALSE;
			}

		else
			{
			// Read what we think is the tag
			lp_new_element->tag_location = p_get_input_location( ap_parse_ctrl ) - 1;

			l_c = p_read_token( ap_parse_ctrl, l_c, lc_tag, sizeof( lc_tag ) );

			if( ! (lp_new_element->p_tag = strdup( lc_tag ) ) )
				return FALSE;

			// Look for =
			while( isspace( l_c ) )
				l_c = p_read( ap_parse_ctrl );

			if( l_c == '=' )
				{
				// We now KNOW that we are looking for a value
				p_parse_value( ap_parse_ctrl, lp_new_element );
				}

			else
				{
				// AMBIGUOUS - What we already have is 
				// -either value without a tag OR tag without a value
				lp_new_element->value_location = lp_new_element->tag_location;
				p_unget_char( ap_parse_ctrl, l_c );
				}
			}


		lp_new_element->tag_number = l_tag_number++;		// See Tag-modify

		lp_last_tag = lp_new_element->p_tag;
		l_last_tag_location = lp_new_element->tag_location;
        }


	ap_parse_ctrl->recursion--;

	return TRUE;
}


//-----------------------------------------------------------------------------
//          Functions for debugging and seeing what is going on
//-----------------------------------------------------------------------------


static void p_display_parse_info( FILE *ah_fout, sumf_message_element *ap_element, int a_indent )
{
    while( ap_element )
        {
        int l_i;
		char *lp_tag = ap_element->p_tag;

		if( ! lp_tag )
			lp_tag = "?";

        for( l_i=0; l_i<a_indent; l_i++ )
            fprintf( ah_fout, "    " );

        fprintf( ah_fout, "Name: %s, tag offset: %d, tag number: %d, value offset: %d\n",
                lp_tag,
                ap_element->tag_location,
				ap_element->tag_number,
                ap_element->value_location );

        if( ap_element->p_compound )
            p_display_parse_info( ah_fout, ap_element->p_compound, a_indent + 1 );

        ap_element = ap_element->p_next;
        }
}


void fumf_display_parse_info( FILE *ah_fout, sumf_parse_ctrl *ap_parse_ctrl )
{
	p_display_parse_info( ah_fout, ap_parse_ctrl->root_location.p_element, 0 );
}

//-----------------------------------------------------------------------------
//          High-level functions for querying message contents
//-----------------------------------------------------------------------------


static eumf_error p_locate_parameter( sumf_location *ap_location,
											const char *ap_tagging,
											int a_instance )
{
    char	lc_tag_looked_for[KUMF_MAX_TAG];
    int		l_i;
	int		l_untagged_number = 0;
	sumf_tag_info *lp_tag_info = ap_location->p_tag_info;


	MUMF_ASSERT( a_instance >= 0 );


	// Sort out untagged indexing first
	if( isdigit( *ap_tagging ) )
		{
		l_untagged_number = *ap_tagging++ - '0';

		while( isdigit( *ap_tagging ) )
			{
			l_untagged_number *= 10;
			l_untagged_number += *ap_tagging++ - '0';
			}
		}

    // Extract the next segment to be looked for
    for( l_i = 0; *ap_tagging != '/' && *ap_tagging != '\0'; l_i++ )
        lc_tag_looked_for[l_i] = *ap_tagging++;

    lc_tag_looked_for[l_i] = '\0';


#ifdef UMF_USE_COMPILED_TAG_INFO
	if( lc_tag_looked_for[0] != '\0' )
		l_untagged_number = (int)(lp_tag_info->p_definition);

	else if( l_untagged_number >= (int)(lp_tag_info->p_definition) )
		return EUMF_PARAMETER_DOES_NOT_EXIST;


	if( lp_tag_info->p_name == gumf_is_struct )
		{
		lp_tag_info++;
		for( l_i=0; l_i < l_untagged_number; l_i++ )
			lp_tag_info++;
		}

	else 
		{
		lp_tag_info++;
		if( lc_tag_looked_for[0] == '\0' )
			{
			if( lp_tag_info->p_name != gumf_is_untagged )
				return EUMF_PARAMETER_DOES_NOT_EXIST;
			}
		else
			{
			if( lp_tag_info->p_name == gumf_is_untagged )
				lp_tag_info++;
			}
		}
#endif


	// Handle the untagged part.  See Locate-Correct
	while( ap_location->p_element && ap_location->p_element->tag_number < l_untagged_number )
		ap_location->p_element = ap_location->p_element->p_next;


	// If we're looking for untagged, find it now
	if( lc_tag_looked_for[0] == '\0' )
		{
        if( *ap_tagging == '/' )
			{
#ifdef UMF_USE_COMPILED_TAG_INFO
			if( !lp_tag_info->p_definition )
				return EUMF_PARAMETER_DOES_NOT_EXIST;

			ap_location->p_tag_info = lp_tag_info->p_definition;
#endif

			ap_location->p_element = ap_location->p_element->p_compound;

            return( p_locate_parameter( ap_location, 
                                        ap_tagging + 1,
                                        a_instance ) );
			}

		else // ( *ap_tagging == '\0' )
			{
			while( ap_location->p_element && ap_location->p_element->tag_number == l_untagged_number )
				{
				if( a_instance-- == 0 )
					{
					ap_location->p_tag_info = lp_tag_info;
					return EUMF_OK;
					}

				ap_location->p_element = ap_location->p_element->p_next;
				}

			return EUMF_PARAMETER_NOT_FOUND;
			}
		}

	else	// ( lc_tag_looked_for[0] != '\0' )
		{
#ifdef UMF_USE_COMPILED_TAG_INFO
		while( lp_tag_info->p_name != gumf_is_struct && lp_tag_info->p_name != gumf_is_union )
			{
			if( lp_tag_info->p_name != gumf_is_untagged &&										// Match tagged
							lc_tag_looked_for[0] == *(lp_tag_info->p_name) &&
							strcmp( &lc_tag_looked_for[1], lp_tag_info->p_name + 1 ) == 0 )
				{
				break;
				}

			lp_tag_info++;
			}

		if( lp_tag_info->p_name == gumf_is_struct || lp_tag_info->p_name == gumf_is_union )
			return EUMF_PARAMETER_DOES_NOT_EXIST;

		if( *ap_tagging == '/' && ! lp_tag_info->p_definition )
				return EUMF_PARAMETER_DOES_NOT_EXIST;
#endif


		// Look for tagged instance
		while( ap_location->p_element )
			{
			if( ap_location->p_element->p_tag &&										// Match tagged
							lc_tag_looked_for[0] == *(ap_location->p_element->p_tag) &&
							strcmp( &lc_tag_looked_for[1], ap_location->p_element->p_tag + 1 ) == 0 )
				{
				if( *ap_tagging == '/' )
					{
#ifdef UMF_USE_COMPILED_TAG_INFO
					ap_location->p_tag_info = lp_tag_info->p_definition;
#endif
					ap_location->p_element = ap_location->p_element->p_compound;

					return( p_locate_parameter( ap_location, 
												ap_tagging + 1,
												a_instance ) );
					}
				else  // ( *ap_tagging == '\0' )
					{
					if( a_instance-- == 0 )
						{
						ap_location->p_tag_info = lp_tag_info;
						return EUMF_OK;
						}
					}

				}

			ap_location->p_element = ap_location->p_element->p_next;
			}
		}


    return EUMF_PARAMETER_NOT_FOUND;
}


static sumf_message_element *p_locate( sumf_parse_ctrl *ap_parse_ctrl, 
										const char *ap_tagging,
										int a_instance )
{
	sumf_location l_location = ap_parse_ctrl->base_location;

	ap_parse_ctrl->error = p_locate_parameter( &l_location, 
												ap_tagging,
												a_instance );

	if( ap_parse_ctrl->error != EUMF_OK )
		return NULL;

	return l_location.p_element;
}


//-----------------------------------------------------------------------------
//                  Helper functions for reading the values
//-----------------------------------------------------------------------------


static tumf_bool p_get_first_char( sumf_parse_ctrl *ap_parse_ctrl,
									tumf_input_location a_location,
									int *ap_c )
{
    p_set_input_location( ap_parse_ctrl, a_location );

    while( (*ap_c = p_read( ap_parse_ctrl )) != KUMF_EOM && isspace( *ap_c ) )
        {
        }


    if( *ap_c == KUMF_EOM )
		{
		ap_parse_ctrl->error = EUMF_MESSAGE_TOO_SHORT;
		return FALSE;
		}

	return TRUE;
}


static tumf_bool p_read_pos_integer( sumf_parse_ctrl *ap_parse_ctrl, int *ap_int )
{
	int l_integer = 0;
	int l_c = p_read( ap_parse_ctrl );

	if( !isdigit( l_c ) )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	do
		{
		l_integer = 10 * l_integer + l_c - '0';

		l_c = p_read( ap_parse_ctrl );

		} while( isdigit( l_c ) );

	p_unget_char( ap_parse_ctrl, l_c );

	*ap_int = l_integer;

	return TRUE;
}


static tumf_bool p_get_ascii_length( sumf_parse_ctrl *ap_parse_ctrl,
										sumf_message_element *ap_element,
										int *ap_length )
{
    int l_error = EUMF_OK;
    int l_c;
    int l_length;


    if( ! p_get_first_char( ap_parse_ctrl, ap_element->value_location, &l_c ) )
		return FALSE;


    if( l_c != '\'' && !isalnum( l_c ) )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


    // Scan in the entire input string taking care not to overflow ac_string
    // Note direct call to p_get_char so that contents of string can not be interpreted
    // -as comments

	if( l_c != '\'' )
		{
		for( l_length = 0; !isspace( l_c ); l_length++ )
			{
			l_c = p_get_char( ap_parse_ctrl );
			}
		}
	else
		{
		int l_escaped = FALSE;
	    l_length = 0;

		while( (l_c = p_get_char( ap_parse_ctrl )) != KUMF_EOM && 
						(l_c != '\'' || l_escaped) )
			{
			if( !l_escaped && l_c == '\\' )
				l_escaped = TRUE;

			else if( l_escaped )
				{
				l_length++;

				l_escaped = FALSE;
				}

			else
				l_length++;
			}
		}


	*ap_length = l_length;


    return l_error;
}


static tumf_bool p_read_ascii( sumf_parse_ctrl *ap_parse_ctrl,
								sumf_message_element *ap_element,
								char ac_string[],
								int a_max_length )
{
    int l_error = EUMF_OK;
    int l_c;
    int l_length;


    if( ! p_get_first_char( ap_parse_ctrl, ap_element->value_location, &l_c ) )
		return FALSE;


    if( l_c != '\'' && ! MUMF_IS_FIRST_TOKEN_CHAR( l_c ) )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


    a_max_length--;     // Ensure space for terminating '\0' 

	if( l_c != '\'' )
		{
		for( l_length = 0; MUMF_IS_TOKEN_CHAR( l_c, 0 ); l_length++ )
			{
			if( l_length >= a_max_length )
				{
				l_error = EUMF_PARAMETER_TOO_LONG;
				break;
				}

			ac_string[l_length] = l_c;
			l_c = p_get_char( ap_parse_ctrl );
			}
		}
	else
		{
		int l_escaped = FALSE;

		l_length = 0;

		while( (l_c = p_get_char( ap_parse_ctrl )) != KUMF_EOM && 
						(l_c != '\'' || l_escaped) )
			{
			if( l_length >= a_max_length )
				{
				l_error = EUMF_PARAMETER_TOO_LONG;
				break;
				}

			if( !l_escaped && l_c == '\\' )
				l_escaped = TRUE;

			else if( l_escaped )
				{
				if( l_c == 'r' )
					ac_string[l_length++] = '\r';
				else if( l_c == 'n' )
					ac_string[l_length++] = '\n';
				else if( l_c == 't' )
					ac_string[l_length++] = '\t';
				else    // Also covers case for l_c == '\'' & '\\'
					ac_string[l_length++] = l_c;

				l_escaped = FALSE;
				}

			else
				ac_string[l_length++] = l_c;
			}
		}

    ac_string[l_length] = '\0';


	if( l_error != EUMF_OK )
		{
		ap_parse_ctrl->error = l_error;
		return FALSE;
		}

    return TRUE;
}


static tumf_bool p_get_unicode_length( sumf_parse_ctrl *ap_parse_ctrl,
										sumf_message_element *ap_element,
										int *ap_length )
{
    int l_c;
	int l_escaped = FALSE;
    int l_length = 0;


    if( ! p_get_first_char( ap_parse_ctrl, ap_element->value_location, &l_c ) )
		return FALSE;


    if( l_c != '"' )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


    // Scan in the entire input string taking care not to overflow ac_string
    // Note direct call to p_get_char so that contents of string can not be interpreted
    // -as comments


	while( (l_c = p_get_unicode_char( ap_parse_ctrl )) != KUMF_EOM )
		{
		if( l_c == '"' && !l_escaped )
			break; 

		if( !l_escaped && l_c == '\\' )
			l_escaped = TRUE;

		else if( l_escaped )
			{
			l_length++;

			l_escaped = FALSE;
			}

		else
			l_length++;
		}


    if( l_c == KUMF_EOM )
		{
		ap_parse_ctrl->error = EUMF_MESSAGE_TOO_SHORT;
		return FALSE;
		}

	*ap_length = l_length;

    return TRUE;
}



static tumf_bool p_read_unicode( sumf_parse_ctrl *ap_parse_ctrl,
									sumf_message_element *ap_element,
									tumf_unicode ac_string[],
									int a_max_length )
{
    int l_c;
	int l_escaped = FALSE;
    int l_i = 0;
    eumf_error l_error = EUMF_OK;


    if( ! p_get_first_char( ap_parse_ctrl, ap_element->value_location, &l_c ) )
		return FALSE;


    if( l_c != '"' )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


    a_max_length--;     // Ensure space for terminating '\0' 


	while( (l_c = p_get_unicode_char( ap_parse_ctrl )) != KUMF_EOM && 
						l_i < a_max_length )
		{
		if( l_c == '"' && !l_escaped )
			break; 

		if( !l_escaped && l_c == '\\' )
			l_escaped = TRUE;

		else if( l_escaped )
			{
			if( l_c == 'r' )
				ac_string[l_i++] = '\r';
			else if( l_c == 'n' )
				ac_string[l_i++] = '\n';
			else if( l_c == 't' )
				ac_string[l_i++] = '\t';
			else    // Also covers case for l_c == '"' & '\\'
				ac_string[l_i++] = l_c;

			l_escaped = FALSE;
			}

		else
			ac_string[l_i++] = l_c;
		}

    ac_string[l_i] = '\0';


    if( l_c != '"' )
        l_error = EUMF_PARAMETER_TOO_LONG;


	if( l_error != EUMF_OK )
		{
		ap_parse_ctrl->error = l_error;
		return FALSE;
		}

    return TRUE;
}


// Base64 table from RFC 1521
//
//           0 A            17 R            34 i            51 z
//           1 B            18 S            35 j            52 0
//           2 C            19 T            36 k            53 1
//           3 D            20 U            37 l            54 2
//           4 E            21 V            38 m            55 3
//           5 F            22 W            39 n            56 4
//           6 G            23 X            40 o            57 5
//           7 H            24 Y            41 p            58 6
//           8 I            25 Z            42 q            59 7
//           9 J            26 a            43 r            60 8
//          10 K            27 b            44 s            61 9
//          11 L            28 c            45 t            62 +
//          12 M            29 d            46 u            63 /
//          13 N            30 e            47 v
//          14 O            31 f            48 w         (pad) =
//          15 P            32 g            49 x
//          16 Q            33 h            50 y

static char p_base64_char_to_6_bits( int a_c )
{
	MUMF_ASSERT( isalnum( l_c ) || l_c == '+' || l_c == '/' );

	if( a_c >= 'A' && a_c <= 'Z' )
		return( a_c - 'A' );
	else if( a_c >= 'a' && a_c <= 'z' )
		return( a_c - 'a' + 26 );
	else if( isdigit( a_c ) )
		return( a_c - '0' + 52 );
	else if( a_c == '+' )
		return 62;
	else if( a_c == '/' )
		return 63;
}


static void p_base64_convert_24bits_to_bytes( char ac_data[], int a_size, int a_accumulated )
{
	if( a_size >= 3 ) ac_data[2] = a_accumulated & 0xff;

	a_accumulated >>= 8;

	if( a_size >= 2 ) ac_data[1] = a_accumulated & 0xff;

	a_accumulated >>= 8;

	if( a_size >= 1 ) ac_data[0] = a_accumulated & 0xff;
}


static tumf_bool p_get_byte_array_length( sumf_parse_ctrl *ap_parse_ctrl,
											sumf_message_element *ap_element,
											int *ap_length )
{
	int l_c;
	int l_number_accumulated;


    if( ! p_get_first_char( ap_parse_ctrl, ap_element->value_location, &l_c ) )
		return FALSE;


    if( l_c != '^' )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


	l_number_accumulated = 0;

	while( (l_c = p_get_char( ap_parse_ctrl )) != KUMF_EOM &&
			( isalnum( l_c ) || l_c == '+' || l_c == '/' ) )
		{
		l_number_accumulated++;
		l_c = p_get_char( ap_parse_ctrl );
		}

	*ap_length = 3 * (l_number_accumulated / 4);	// 4 chars -> 3 bytes in Base64

	if( l_number_accumulated & 0x03 == 2 )
		*ap_length++;

	else if( l_number_accumulated & 0x03 == 3 )
		*ap_length += 2;


    return TRUE;
}


static tumf_bool p_read_byte_array( sumf_parse_ctrl *ap_parse_ctrl,
										sumf_message_element *ap_element,
										char ac_data[],
										int a_max_length,
										int *ap_result_length )
{
    int l_c;
    int l_i = 0;
	int l_accumulated = 0;
	int l_number_accumulated = 0;


    if( ! p_get_first_char( ap_parse_ctrl, ap_element->value_location, &l_c ) )
		return FALSE;

	if( l_c != '^' )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	l_c = p_get_char( ap_parse_ctrl );

	while( l_i<a_max_length && (isalnum( l_c ) || l_c == '+' || l_c == '/') )
		{
		l_accumulated <<= 6;
		l_accumulated |= p_base64_char_to_6_bits( l_c );
		if( ++l_number_accumulated == 4 )
			{
			p_base64_convert_24bits_to_bytes( &(ac_data[l_i]), a_max_length - l_i, l_accumulated );

			l_accumulated = l_number_accumulated = 0;
			l_i += 3;
			}
		l_c = p_get_char( ap_parse_ctrl );
		}


	if( l_number_accumulated == 1 )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	else if( l_number_accumulated >= 2 )	// Can only be 2 or 3
		{
		int l_residual_bytes = 1;

		if( l_number_accumulated == 3 ) l_residual_bytes = 2;

		while( l_c == '=' && l_number_accumulated++ < 4 )
			{
			l_accumulated <<= 6;
			l_c = p_get_char( ap_parse_ctrl );
			}

		p_base64_convert_24bits_to_bytes( &(ac_data[l_i]), a_max_length - l_i, l_accumulated );

		l_i += l_residual_bytes;
		}


	if( !isspace( l_c ) && l_c != KUMF_EOM )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


	*ap_result_length = l_i;

    return TRUE;
}


static tumf_bool p_get_embedded_length( sumf_parse_ctrl *ap_parse_ctrl,
								sumf_message_element *ap_element,
								int *ap_result_length )
{
    int l_error = EUMF_OK;
    int l_c;
    int l_length = 0;
	int l_escaped = FALSE;


    if( ! p_get_first_char( ap_parse_ctrl, ap_element->value_location, &l_c ) )
		return FALSE;


    if( l_c != '(' )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


	l_length = 0;

	while( (l_c = p_get_char( ap_parse_ctrl )) != KUMF_EOM )
		{
		if( l_c == ')' && !l_escaped )
			break; 

		if( !l_escaped && l_c == '\\' )
			l_escaped = TRUE;

		else if( l_escaped )
			{
			l_length++;

			l_escaped = FALSE;
			}

		else
			l_length++;
		}


	*ap_result_length = l_length;


	if( l_error != EUMF_OK )
		{
		ap_parse_ctrl->error = l_error;
		return FALSE;
		}

    return TRUE;
}


static tumf_bool p_read_embedded( sumf_parse_ctrl *ap_parse_ctrl,
								sumf_message_element *ap_element,
								char ac_data[],
								int a_max_length,
								int *ap_result_length )
{
    int l_c;
    int l_i = 0;
	int l_escaped = FALSE;


    if( ! p_get_first_char( ap_parse_ctrl, ap_element->value_location, &l_c ) )
		return FALSE;


    if( l_c != '(' )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


	l_i = 0;

	while( (l_c = p_get_char( ap_parse_ctrl )) != KUMF_EOM && 
						l_i < a_max_length )
		{
		if( l_c == ')' && !l_escaped )
			break; 

		if( !l_escaped && l_c == '\\' )
			l_escaped = TRUE;

		else if( l_escaped )
			{
			ac_data[l_i++] = l_c; // Covers case for l_c == '\\' & '\)'

			l_escaped = FALSE;
			}

		else
			ac_data[l_i++] = l_c;
		}


	*ap_result_length = l_i;


    if( l_c != ')' )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_TOO_LONG;
		return FALSE;
		}


    return TRUE;
}


static void p_free_element_info( sumf_message_element *ap_element )
{
    sumf_message_element *lp_element_to_free;

    while( ap_element )
        {
        if( ap_element->p_tag )
			free( ap_element->p_tag );

        if( ap_element->p_compound )
            p_free_element_info( ap_element->p_compound );

        lp_element_to_free = ap_element;

        ap_element = ap_element->p_next;

        free( lp_element_to_free );
        }
}


//-----------------------------------------------------------------------------
//                          Start of API functions
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//              API functions for initiating the parse routines
//-----------------------------------------------------------------------------


// Func name:   fumf_is_message_complete
//------------------------------------------------------------------------------
// Parameters:	const char *ap_message - Pointer to the buffer containing the
//							message. 
//				int a_length - The length of the buffer containg the message.
// Returns:		int - The length of the message if complete, or -1 if 
//							incomplete.
// Description: Test whether the message is complete according to the method
//				of having an unmatched closing brace at the end of the message.

int fumf_is_message_complete( const char *ap_message, int a_length )
{
	int l_c;
	sumf_parse_ctrl l_parse_ctrl;
	int l_nesting = 0;

	fumf_initialise( &l_parse_ctrl );

	l_parse_ctrl.p_message = l_parse_ctrl.p_next = ap_message;
	l_parse_ctrl.p_message_end = ap_message + a_length;


	l_c = p_read( &l_parse_ctrl );


	while( l_c != KUMF_EOM )
		{
		if( l_c == '{' )
			l_nesting++;

		else if( l_c == '}' )
			{
			l_nesting--;
			if( l_nesting < 0 )
				return l_parse_ctrl.p_next - l_parse_ctrl.p_message;
			}

		else if( l_c == '=' || isspace( l_c ) )
			{
			// Ignore and move on
			}

		else
			p_skip_value( &l_parse_ctrl, l_c );		// Skip tags as well as values

		
		l_c = p_read( &l_parse_ctrl );
		}


	return -1;
}


// Func name:   fumf_initialise
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise
//
// Description: Initialises the user allocated sumf_parse_ctrl structure with
//				the details of the message to be parsed.  
//
//				N.B. This function is called as part of fumf_parse_message and
//				so it is not normally necessary to call it separately.

tumf_bool fumf_initialise( sumf_parse_ctrl *ap_parse_ctrl )
{
	ap_parse_ctrl->allocation_failed = FALSE;
	ap_parse_ctrl->root_location.p_element = NULL;
	ap_parse_ctrl->base_location.p_element = NULL;
	ap_parse_ctrl->base_location.p_tag_info = 
			ap_parse_ctrl->root_location.p_tag_info;	// Allow for root to be set already
	ap_parse_ctrl->unget_char = '\0';
	ap_parse_ctrl->recursion = 0;

	ap_parse_ctrl->error = EUMF_OK;

	return TRUE;
}


// Func name:   fumf_set_tag_info_root
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				sumf_tag_info *ap_tag_info - Pointer to the beginning of the
//							structure containing the tagging information.
// Returns:		void
//
// Description: Specifies the base address of the tag info data structure.

void fumf_set_tag_info_root( sumf_parse_ctrl *ap_parse_ctrl, sumf_tag_info *ap_tag_info )
{
	ap_parse_ctrl->root_location.p_tag_info = 
						ap_parse_ctrl->base_location.p_tag_info = ap_tag_info;
}


// Func name:   fumf_get_last_error
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
// Returns:		eumf_error - The last error code.
//
// Description: Returns the lastest error and resets the error condition to
//				EUMF_OK.

eumf_error fumf_get_last_error( sumf_parse_ctrl *ap_parse_ctrl )
{
	eumf_error l_last_error = ap_parse_ctrl->error;

	ap_parse_ctrl->error = EUMF_OK;

	return l_last_error;
}


// Func name:   fumf_parse_message
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				const char *ap_message - Pointer to the encoded message.
//				int a_length - The length of the encoded message.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Parses the message.  This builds a tree representation of the
//				message that the parameter reading functions use to read parts
//				of the message.
//
//				fumf_release should be called to release the memory allocated 
//				by this function once all the interaction with the message has
//				been completed.

tumf_bool fumf_parse_message( sumf_parse_ctrl *ap_parse_ctrl, 
								const char *ap_message, 
								int a_length )
{
	int l_parse_result;

	fumf_initialise( ap_parse_ctrl );

	ap_parse_ctrl->p_message = ap_parse_ctrl->p_next = ap_message;
	ap_parse_ctrl->p_message_end = ap_message + a_length;

    ap_parse_ctrl->allocation_failed = FALSE;

	l_parse_result = p_parse( ap_parse_ctrl, &(ap_parse_ctrl->base_location.p_element) );

	ap_parse_ctrl->root_location.p_element = ap_parse_ctrl->base_location.p_element;

	if( l_parse_result == FALSE )
		{
		if( ap_parse_ctrl->allocation_failed && ap_parse_ctrl->error == EUMF_OK )
			ap_parse_ctrl->error = EUMF_ALLOCATION_FAILED;

		return FALSE;
		}

    return TRUE;
}


//-----------------------------------------------------------------------------
//        API functions for setting the read position within messages
//-----------------------------------------------------------------------------


// Func name:   fumf_set_location
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				sumf_location *ap_old_location - Where the previous location 
//							information should be stored.  N.B. This can be
//							NULL if the location information is not required.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Sets the base read location in the tree from which subsequent
//				parameter specifications are made.  If / is the first character
//				then the specification is addressed from the root of the 
//				message tree, otherwise it is addressed from the current
//				location.  By analogy to file directories, it is similar to the 
//				cd command.
// Returns:		The previously set read location

tumf_bool fumf_set_location( sumf_parse_ctrl *ap_parse_ctrl,
									const char *ap_tagging, 
									int a_instance,
									sumf_location *ap_old_location )
{
	if( ap_old_location )
		*ap_old_location = ap_parse_ctrl->base_location;

	
	if( *ap_tagging == '/' )
		{
		ap_parse_ctrl->base_location = ap_parse_ctrl->root_location;
		ap_tagging++;
		}


	if( *ap_tagging != '\0' )
		{
		sumf_location l_location = ap_parse_ctrl->base_location;

		ap_parse_ctrl->error = p_locate_parameter( &l_location, 
													ap_tagging,
													a_instance );

		if( ap_parse_ctrl->error != EUMF_OK )
			return FALSE;

		else if( l_location.p_element == NULL || l_location.p_element->p_compound == NULL )
			{
			ap_parse_ctrl->error = EUMF_PARAMETER_NOT_FOUND;
			return FALSE;
			}

		ap_parse_ctrl->base_location.p_element = l_location.p_element->p_compound;
		ap_parse_ctrl->base_location.p_tag_info = l_location.p_tag_info->p_definition;
		}


	return TRUE;
}


// Func name:   fumf_restore_location
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				sumf_location a_base - The base location information being 
//							restored.
// Returns:		sumf_location - The previous base location.
//
// Description: Restores the read location to a previous location returned 
//				from any of the other location setting functions.

sumf_location fumf_restore_location( sumf_parse_ctrl *ap_parse_ctrl,
										sumf_location a_base )
{
	sumf_location l_old_location = ap_parse_ctrl->base_location;

	ap_parse_ctrl->base_location = a_base;

	return l_old_location;
}


//-----------------------------------------------------------------------------
//                   API functions for reading the values
//-----------------------------------------------------------------------------


// Func name:   fumf_is_present
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Indicates whether the named parameter is present.

tumf_bool fumf_is_present( sumf_parse_ctrl *ap_parse_ctrl,
							const char ac_tagging[],
							int a_instance )
{
    sumf_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );

    if( lp_element )
        return TRUE;

    return FALSE;
}


// Func name:   fumf_read_void
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Reads the void value associated with the named parameter.  In 
//				practice, this just checks for the presence of the void's tag.

tumf_bool fumf_read_void( sumf_parse_ctrl *ap_parse_ctrl,
						const char ac_tagging[],
						int a_instance )
{
	if( !fumf_is_present( ap_parse_ctrl, ac_tagging, a_instance ) )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_NOT_FOUND;
		return FALSE;
		}

	return TRUE;
}


// Func name:   fumf_read_boolean
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int *ap_boolean - Pointer to where the result is to be placed,
//							which will either be TRUE or FALSE.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Reads the boolean value associated with the named parameter.

tumf_bool fumf_read_boolean( sumf_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								int *ap_boolean )
{
    int l_c;
    sumf_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );


    if( ! lp_element )
		return FALSE;

	if( ! p_get_first_char( ap_parse_ctrl, lp_element->value_location, &l_c ) )
		return FALSE;

    if( l_c == 'T' || l_c == 't' )
        *ap_boolean = TRUE;

    else if( l_c == 'F' || l_c == 'f' )
        *ap_boolean = FALSE;

    else
		{
        ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

    return TRUE;
}


// Func name:   fumf_read_integer
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int *ap_integer - Pointer to where the integer result is to be 
//							placed.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Reads the integer value associated with the named parameter.

tumf_bool fumf_read_integer( sumf_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								int *ap_integer )
{
    int l_c;
    int l_value = 0;
    int l_positive = TRUE;
    sumf_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );


    if( ! lp_element )
		return FALSE;


	if( ! p_get_first_char( ap_parse_ctrl, lp_element->value_location, &l_c ) )
		return FALSE;

    if( l_c == '-' )
        {
        l_positive = FALSE;

        l_c = p_read( ap_parse_ctrl );
        }

    if( !isdigit( l_c ) )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

    else
        {
        while( isdigit( l_c ) )
            {
            l_value *= 10;

            l_value += l_c - '0';

            l_c = p_read( ap_parse_ctrl );
            }

        if( l_positive )
            *ap_integer = l_value;
        else
            *ap_integer = -l_value;
        }


    return TRUE;
}


// Func name:   fumf_read_ipv4
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				unsigned char ac_addr[4] - Where the 4 bytes of the IPv4 
//							network address are to be placed.  This is in
//							network byte order.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Reads the IPv4 address value associated with the named 
//				parameter.

tumf_bool fumf_read_ipv4( sumf_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								unsigned char ac_addr[4] )
{
	int l_int1, l_int2, l_int3, l_int4;
    sumf_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );

    if( ! lp_element )
		return FALSE;


    p_set_input_location( ap_parse_ctrl, lp_element->value_location );

	if( ! p_read_pos_integer( ap_parse_ctrl, &l_int1 ) ||
					'.' != p_read( ap_parse_ctrl ) ||
				! p_read_pos_integer( ap_parse_ctrl, &l_int2 ) ||
					'.' != p_read( ap_parse_ctrl ) ||
				! p_read_pos_integer( ap_parse_ctrl, &l_int3 ) ||
					'.' != p_read( ap_parse_ctrl ) ||
				! p_read_pos_integer( ap_parse_ctrl, &l_int4 ) )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	if( l_int1 < 0 || l_int1 > 255 ||
				l_int2 < 0 || l_int2 > 255 ||
				l_int3 < 0 || l_int3 > 255 ||
				l_int4 < 0 || l_int4 > 255 )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	ac_addr[0] = (unsigned char)l_int1;
	ac_addr[1] = (unsigned char)l_int2;
	ac_addr[2] = (unsigned char)l_int3;
	ac_addr[3] = (unsigned char)l_int4;


    return TRUE;
}


// Func name:   fumf_read_date
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				tumf_date *ap_date - Pointer to where the date is to be placed.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Reads the date value associated with the named 
//				parameter.

tumf_bool fumf_read_date( sumf_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								tumf_date *ap_date )
{
    sumf_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );

    if( ! lp_element )
		return FALSE;

    p_set_input_location( ap_parse_ctrl, lp_element->value_location );

	if( ! p_read_pos_integer( ap_parse_ctrl, &(ap_date->year) ) ||
					'-' != p_read( ap_parse_ctrl ) ||
				! p_read_pos_integer( ap_parse_ctrl, &(ap_date->month) ) ||
					'-' != p_read( ap_parse_ctrl ) ||
				! p_read_pos_integer( ap_parse_ctrl, &(ap_date->day) ) )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	if( ap_date->month < 1 || ap_date->month > 12 ||
				ap_date->day < 1 || ap_date->day > 31 )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

    return TRUE;
}


// Func name:   fumf_read_time
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				tumf_date *ap_time - Pointer to where the time is to be placed.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Reads the date value associated with the named 
//				parameter.

tumf_bool fumf_read_time( sumf_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								tumf_time *ap_time )
{
	int l_c;
    sumf_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );

    if( ! lp_element )
		return FALSE;


    p_set_input_location( ap_parse_ctrl, lp_element->value_location );

	if( ! p_read_pos_integer( ap_parse_ctrl, &(ap_time->hour) ) ||
					':' != p_read( ap_parse_ctrl ) ||
				! p_read_pos_integer( ap_parse_ctrl, &(ap_time->min) ) )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	if( ap_time->hour < 0 || ap_time->hour > 23 ||
				ap_time->min < 0 || ap_time->min > 59 )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}
					
	ap_time->sec = 0;

	l_c = p_read( ap_parse_ctrl );

	if( isspace( l_c ) )
		return TRUE;

	if( ':' != l_c ||
				! p_read_pos_integer( ap_parse_ctrl, &(ap_time->sec) ) )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	if( ap_time->sec < 0 || ap_time->sec > 59 )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	l_c = p_read( ap_parse_ctrl );

	if( ! isspace( l_c ) )
		{
		ap_parse_ctrl->error = EUMF_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


    return TRUE;
}


// Func name:   fumf_read_ascii
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				char ac_string[] - Where the ASCII string is to be placed.
//				int a_max_length - The length of the ac_string array.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Reads the unquoted or quoted ascii string value associated with 
//				the named parameter.

tumf_bool fumf_read_ascii( sumf_parse_ctrl *ap_parse_ctrl,
							const char ac_tagging[],
							int a_instance, 
							char ac_string[],
							int a_max_length )
{
    sumf_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );


    if( ! lp_element )
		return FALSE;

	return p_read_ascii( ap_parse_ctrl, lp_element, ac_string, a_max_length );
}


// Func name:   fumf_read_ascii_dynamic
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				char **app_string - Pointer to where the pointer to the 
//							dynamically allocated string buffer is to be 
//							placed.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Reads the unquoted or quoted ascii string value associated with 
//				the named parameter into an array that is dynamically allocated
//				by the library.  
//
//				It is the responsibility of the calling code to free this memory
//				when the information is no longer required.

tumf_bool fumf_read_ascii_dynamic( sumf_parse_ctrl *ap_parse_ctrl,
										const char ac_tagging[],
										int a_instance, 
										char **app_string )
{
	int l_error;
	int l_length;
    sumf_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );


    if( ! lp_element )
		return FALSE;


	*app_string = NULL;


	l_error = p_get_ascii_length( ap_parse_ctrl, lp_element, &l_length );

	if( l_error != EUMF_OK )
		{
		ap_parse_ctrl->error = l_error;
		return FALSE;
		}

	
	l_length++;	// Allow for terminating \0

	*app_string = (char *)fumf_alloc( ap_parse_ctrl, l_length * sizeof(char) );

	if( *app_string == NULL )
		{
		ap_parse_ctrl->error = EUMF_ALLOCATION_FAILED;
		return FALSE;
		}


	return p_read_ascii( ap_parse_ctrl, lp_element, *app_string, l_length );
}


// Func name:   fumf_read_unicode
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				tumf_unicode ac_string[] - Where the Unicode string is to be 
//							placed.
//				int a_max_length - The length of the ac_string array.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Reads the unicode string value associated with the named 
//				parameter.

tumf_bool fumf_read_unicode( sumf_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								tumf_unicode ac_string[],
								int a_max_length )
{
    sumf_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );


    if( ! lp_element )
		return FALSE;


	return p_read_unicode( ap_parse_ctrl, lp_element, ac_string, a_max_length );
}


// Func name:   fumf_read_unicode_dynamic
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				tumf_unicode **app_string - Pointer to where the pointer to the 
//							dynamically allocated string buffer is to be 
//							placed.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Reads the unicode string value associated with the named 
//				parameter into an array that is dynamically allocated by the 
//				library.  
//
//				It is the responsibility of the calling code to free this memory
//				when the information is no longer required.

tumf_bool fumf_read_unicode_dynamic( sumf_parse_ctrl *ap_parse_ctrl,
										const char ac_tagging[],
										int a_instance, 
										tumf_unicode **app_string )
{
	int l_length;
	int l_error;
    sumf_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );


    if( ! lp_element )
		return FALSE;


	*app_string = NULL;


	l_error = p_get_unicode_length( ap_parse_ctrl, lp_element, &l_length );

	if( l_error != EUMF_OK )
		{
		ap_parse_ctrl->error = l_error;
		return FALSE;
		}
	
	l_length++;	// Allow for terminating \0

	*app_string = (tumf_unicode *)fumf_alloc( ap_parse_ctrl, l_length * sizeof(tumf_unicode) );

	if( *app_string == NULL )
		{
		ap_parse_ctrl->error = EUMF_ALLOCATION_FAILED;
		return FALSE;
		}


	return p_read_unicode( ap_parse_ctrl, lp_element, *app_string, l_length );
}


// Func name:   fumf_read_const
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				char ac_string[] - Where the const ASCII string is to be placed.
//				int a_max_length - The length of the ac_string array.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Reads the const value associated with the named parameter.

tumf_bool fumf_read_const( sumf_parse_ctrl *ap_parse_ctrl,
							const char ac_tagging[],
							int a_instance, 
							char ac_string[],
							int a_max_length )
{
	return fumf_read_ascii( ap_parse_ctrl,
								ac_tagging,
								a_instance, 
								ac_string,
								a_max_length );
}


// Func name:   fumf_byte_array
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				char ac_data[] - Where the bytes array data is to be placed.
//				int a_max_length - The length of the ac_data array.
//				int *ap_result_length - Where the resulting number of bytes 
//							stored in ac_data should be stored.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Reads the byte array value associated with the named parameter.

tumf_bool fumf_read_byte_array( sumf_parse_ctrl *ap_parse_ctrl,
							const char ac_tagging[],
							int a_instance, 
							char ac_data[],
							int a_max_length,
							int *ap_result_length )
{
    sumf_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );


    if( ! lp_element )
		return FALSE;


	return p_read_byte_array( ap_parse_ctrl, lp_element, ac_data, a_max_length, ap_result_length );
}


// Func name:   fumf_read_byte_array_dynamic
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				char **app_data - Pointer to where the pointer to the 
//							dynamically allocated data buffer is to be 
//							placed.
//				int *ap_result_length - Where the resulting number of bytes 
//							stored in app_data should be stored.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Reads the byte array value associated with the named parameter 
//				into an array that is dynamically allocated by the library.
//
//				It is the responsibility of the calling code to free this memory
//				when the information is no longer required.

tumf_bool fumf_read_byte_array_dynamic( sumf_parse_ctrl *ap_parse_ctrl,
										const char ac_tagging[],
										int a_instance, 
										char **app_data,
										int *ap_result_length )
{
	int l_error;
	int l_length;
    sumf_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );

    if( ! lp_element )
		return FALSE;


	*app_data = NULL;


	l_error = p_get_byte_array_length( ap_parse_ctrl, lp_element, &l_length );

	if( l_error != EUMF_OK )
		{
		ap_parse_ctrl->error = l_error;
		return FALSE;
		}

	
	*app_data = (char *)fumf_alloc( ap_parse_ctrl, l_length * sizeof(char) );

	if( *app_data == NULL )
		{
		ap_parse_ctrl->error = EUMF_ALLOCATION_FAILED;
		return FALSE;
		}


	return p_read_byte_array( ap_parse_ctrl, lp_element, *app_data, l_length, ap_result_length );
}


// Func name:   fumf_read_embedded
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				char ac_data[] - Where the bytes array data is to be placed.
//				int a_max_length - The length of the ac_data array.
//				int *ap_result_length - Where the resulting number of bytes 
//							stored in ac_data should be stored.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Reads the embedded value associated with the named parameter.

tumf_bool fumf_read_embedded( sumf_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								char ac_data[],
								int a_max_length,
								int *ap_result_length )
{
    sumf_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );


    if( ! lp_element )
		return FALSE;


	return p_read_embedded( ap_parse_ctrl, lp_element, ac_data, a_max_length, ap_result_length );
}


// Func name:   fumf_read_embedded_dynamic
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				char **app_data - Pointer to where the pointer to the 
//							dynamically allocated data buffer is to be 
//							placed.
//				int *ap_result_length - Where the resulting number of bytes 
//							stored in app_data should be stored.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Reads the embedded value associated with the named 
//				parameter into an array that is dynamically allocated by the 
//				library.  
//
//				It is the responsibility of the calling code to free this memory
//				when the information is no longer required.

tumf_bool fumf_read_embedded_dynamic( sumf_parse_ctrl *ap_parse_ctrl,
										const char ac_tagging[],
										int a_instance, 
										char **app_data,
										int *ap_result_length )
{
	int l_error;
	int l_length;
    sumf_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );

    if( lp_element == NULL )
		return FALSE;


	*app_data = NULL;


	l_error = p_get_byte_array_length( ap_parse_ctrl, lp_element, &l_length );

	if( l_error != EUMF_OK )
		{
		ap_parse_ctrl->error = l_error;
		return FALSE;
		}

	
	*app_data = (char *)fumf_alloc( ap_parse_ctrl, l_length * sizeof(char) );

	if( *app_data == NULL )
		{
		ap_parse_ctrl->error = EUMF_ALLOCATION_FAILED;
		return FALSE;
		}


	return p_read_embedded( ap_parse_ctrl, lp_element, *app_data, l_length, ap_result_length );
}


// Func name:   fumf_read_union
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				char ac_string[] - Where the string specified in the union is 
//							to be placed.
//				int a_max_length - The length of the ac_string array.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Reads the selected option of the union associated with the 
//				named parameter.

tumf_bool fumf_read_union( sumf_parse_ctrl *ap_parse_ctrl,
									const char ac_tagging[],
									int a_instance, 
									char ac_string[],
									int a_max_length )
{
	return fumf_read_ascii( ap_parse_ctrl,
								ac_tagging,
								a_instance, 
								ac_string,
								a_max_length );
}


// Func name:   fumf_read_enumerated_union
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				sumf_tag_translation ac_tag_translation[] - Pointer to the 
//							array mapping between ascii words and integer 
//							values to be used for the enumeration.
// Returns:		int - The integer valid corresponding to the union value 
//							present.  If there is not an entry in the 
//							translation table, then -1 is returned.
//
// Description: Reads the selected option of the union associated with the 
//				named parameter and converts it to an integer value according to
//				the contents of the array ac_tag_translation.  This is indended
//				for use in switch statements.

int fumf_read_enumerated_union( sumf_parse_ctrl *ap_parse_ctrl,
									const char ac_tagging[],
									int a_instance, 
									sumf_tag_translation ac_tag_translation[] )
{
	char lc_union_option[KUMF_MAX_TAG];

	if( fumf_read_union( ap_parse_ctrl,  
							ac_tagging, a_instance, 
							lc_union_option, sizeof(lc_union_option) ) )
		{
		int l_index;
		char *lp_sought = lc_union_option;

		if( isdigit( lc_union_option[0] ) ) 
			lp_sought = "";

		for( l_index=0; ac_tag_translation[l_index].text_tag != NULL; l_index++ )
			{
			if( ac_tag_translation[l_index].text_tag[0] == lp_sought[0] &&
					( lp_sought[0] == '\0' ||
						strcmp( &(ac_tag_translation[l_index].text_tag[1]), 
									&(lp_sought[1]) ) == 0 ) )
				{
				return( ac_tag_translation[l_index].numerical_tag );
				}
			}

		}

	return -1;
}


//-----------------------------------------------------------------------------
//                         Functions for clearing up
//-----------------------------------------------------------------------------


// Func name:   fumf_release
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
// Returns:		void
//
// Description: Releases all the resources associated with parsing the message.
//				This function must be called when the code have finished with
//				a parsed message to avoid memory leaks. 

void fumf_release( sumf_parse_ctrl *ap_parse_ctrl )
{
	p_free_element_info( ap_parse_ctrl->root_location.p_element );

	ap_parse_ctrl->root_location.p_element = 
					ap_parse_ctrl->base_location.p_element = NULL;
}


//-----------------------------------------------------------------------------
//                      Functions to support encoding
//-----------------------------------------------------------------------------

//   UCS-2 range (hex.)		UTF-8 octet sequence (binary)
//   0000-007F				0xxxxxxx
//   0080-07FF				110xxxxx 10xxxxxx
//   0800-FFFF				1110xxxx 10xxxxxx 10xxxxxx

static int p_unicode_to_escaped_utf8( char ac_utf8[3], tumf_unicode a_unicode_char )
{
	if( a_unicode_char == '\\' || a_unicode_char == '"' )
		{
		ac_utf8[0] = '\\';
		ac_utf8[1] = (char)a_unicode_char;
		return 2;
		}
	else if( a_unicode_char <= 0x7f )
		{
		ac_utf8[0] = (char)a_unicode_char;
		return 1;
		}

	else if( a_unicode_char <= 0x7ff )
		{
		// 2 bytes
		ac_utf8[1] = ((char)a_unicode_char & 0x3f) | 0x80;
		a_unicode_char >>= 6;
		ac_utf8[0] = ((char)a_unicode_char & 0x1f) | 0xC0;
		return 2;
		}

	else
		{
		// 3 bytes
		ac_utf8[2] = ((char)a_unicode_char & 0x3f) | 0x80;
		a_unicode_char >>= 6;
		ac_utf8[1] = ((char)a_unicode_char & 0x3f) | 0x80;
		a_unicode_char >>= 6;
		ac_utf8[0] = ((char)a_unicode_char & 0x0f) | 0xE0;
		return 3;
		}
}


static char p_base64_6_bits_to_base64_char( int a_c )
{
	MUMF_ASSERT( a_c <= 0x00ffffff );

	if( a_c >= 0 && a_c <= 25 )
		return( a_c + 'A' );
	else if( a_c >= 26 && a_c <= 51 )
		return( a_c - 26 + 'a' );
	else if( a_c >= 52 && a_c <= 61 )
		return( a_c - 52 + '0' );
	else if( a_c == 62 )
		return '+';
	else if( a_c == 63 )
		return '/';
}


static void p_base64_convert_24bits_to_4x6bits( char ac_data[], int a_accumulated )
{
	ac_data[3] = p_base64_6_bits_to_base64_char( a_accumulated & 0x3f );

	a_accumulated >>= 6;

	ac_data[2] = p_base64_6_bits_to_base64_char( a_accumulated & 0x3f );

	a_accumulated >>= 6;

	ac_data[1] = p_base64_6_bits_to_base64_char( a_accumulated & 0x3f );

	a_accumulated >>= 6;

	ac_data[0] = p_base64_6_bits_to_base64_char( a_accumulated & 0x3f );
}


// Func name:   fumf_enc_get_encoded_ascii_length
//------------------------------------------------------------------------------
// Parameters:	const char ac_ascii[] - Array on which to calculate the 
//							encoded length.
// Returns:		int - The length of the item plus 1 to allow for a terminating
//							\0.
//
// Description: Returns the number of bytes needed to store an escaped ascii
//				string. 
//
//				N.B. The returned length includes space for a terminating \0 
//				so that if a buffer of the returned length is used to store 
//				the encoded version, it can be manipulated as if it were a
//				string using functions such as printf and strcat.

int fumf_enc_get_encoded_ascii_length( const char ac_ascii[] )
{
	int l_ascii_length = 0;
	int l_i;

	for( l_i=0; ac_ascii[l_i] != 0; l_i++ )
		if( ac_ascii[l_i] == '\r' || 
				ac_ascii[l_i] == '\n' || 
				ac_ascii[l_i] == '\\' ||
				ac_ascii[l_i] == '\'' )
			l_ascii_length += 2;
		else
			l_ascii_length++;

	l_ascii_length++;	// Allow for terminating \0

	return l_ascii_length;
}


// Func name:   fumf_enc_generate_encoded_ascii
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_escaped_ascii[] - Where the encoded ASCII string is to
//							be placed.
//				int a_escaped_ascii_max_length - The maximum size of the 
//							ac_escaped_ascii buffer.
//				const char ac_ascii[] - The ASCII string to be encoded.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Converts an ascii string into the escaped format for
//				outputing on the wire. 
//
//				N.B. If successful, a \0 is appended to the end of the string
//				so that the result can be manipulated as a string by 
//				functions such as printf and strcat.

tumf_bool fumf_enc_generate_encoded_ascii( sumf_parse_ctrl *ap_parse_ctrl,
											char ac_escaped_ascii[], 
											int a_escaped_ascii_max_length,
											const char ac_ascii[] )
{
	int l_i;
	int l_escaped_ascii_length = 0;

	// Allow space for termination char or \ escape char
	a_escaped_ascii_max_length--;

	for( l_i=0; ac_ascii[l_i] != 0 && 
					l_escaped_ascii_length < a_escaped_ascii_max_length; 
				l_i++ )
		{
		if( ac_ascii[l_i] == '\r' ) 
			{
			ac_escaped_ascii[l_escaped_ascii_length++] = '\\';
			ac_escaped_ascii[l_escaped_ascii_length++] = 'r';
			}
		else if( ac_ascii[l_i] == '\n' )
			{
			ac_escaped_ascii[l_escaped_ascii_length++] = '\\';
			ac_escaped_ascii[l_escaped_ascii_length++] = 'n';
			}
		else if( ac_ascii[l_i] == '\\' )
			{
			ac_escaped_ascii[l_escaped_ascii_length++] = '\\';
			ac_escaped_ascii[l_escaped_ascii_length++] = '\\';
			}
		else if( ac_ascii[l_i] == '\'' )
			{
			ac_escaped_ascii[l_escaped_ascii_length++] = '\\';
			ac_escaped_ascii[l_escaped_ascii_length++] = '\'';
			}
		else
			ac_escaped_ascii[l_escaped_ascii_length++] = ac_ascii[l_i];
		}


	if( l_escaped_ascii_length > a_escaped_ascii_max_length )
		{
		ap_parse_ctrl->error = EUMF_OUTPUT_TOO_SMALL;
		return FALSE;
		}

	ac_escaped_ascii[l_escaped_ascii_length++] = '\0';

	return TRUE;
}


// Func name:   fumf_enc_get_encoded_unicode_length
//------------------------------------------------------------------------------
// Parameters:	const tumf_unicode ac_unicode[] - Array on which to calculate 
//							the encoded length.
// Returns:		int - The length of the item plus 1 to allow for a terminating
//							\0.
//
// Description: Returns the number of bytes needed to store an encoded Unicode
//				string. 
//
//				N.B. The returned length includes space for a terminating \0 
//				so that if a buffer of the returned length is used to store 
//				the encoded version, it can be manipulated as if it were a
//				string using functions such as printf and strcat.

int fumf_enc_get_encoded_unicode_length( const tumf_unicode ac_unicode[] )
{
	int l_utf8_length = 0;
	char lc_unused[3];
	int l_i;

	for( l_i=0; ac_unicode[l_i] != 0; l_i++ )
		l_utf8_length += p_unicode_to_escaped_utf8( lc_unused, ac_unicode[l_i] );

	l_utf8_length++;	// Allow for terminating \0

	return l_utf8_length;
}


// Func name:   fumf_enc_generate_encoded_unicode
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_utf8[] - Where the encoded byte Unicode string is to
//							be placed.
//				int a_utf8_max_length - The maximum size of the 
//							ac_utf8 buffer.
//				const tumf_unicode ac_unicode[] - The Unicode string to be 
//							encoded.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Converts a Unicode string into the escaped UTF-8 format for
//				outputing on the wire. 
//
//				N.B. If successful, a \0 is appended to the end of the string
//				so that the result can be manipulated as a string by 
//				functions such as printf and strcat.

tumf_bool fumf_enc_generate_encoded_unicode( sumf_parse_ctrl *ap_parse_ctrl,
											char ac_utf8[], 
											int a_utf8_max_length,
											const tumf_unicode ac_unicode[] )
{
	int l_total_utf8_length = 0;
	char lc_converted_char[3];
	int l_utf8_length;
	int l_i;
	int l_j;

	a_utf8_max_length--;		// Allow for terminating \0

	for( l_i=0; ac_unicode[l_i] != 0; l_i++ )
		{
		l_utf8_length = p_unicode_to_escaped_utf8( lc_converted_char, ac_unicode[l_i] );
		if( l_total_utf8_length + l_utf8_length > a_utf8_max_length )
			{
			for( l_j=0; l_j<l_utf8_length; l_j++ )
				ac_utf8[l_total_utf8_length++] = lc_converted_char[l_j]; 
			}
		else
			{
			ap_parse_ctrl->error = EUMF_OUTPUT_TOO_SMALL;
			return FALSE;
			}
		}

	return TRUE;
}


// Func name:   fumf_enc_get_encoded_byte_array_length
//------------------------------------------------------------------------------
// Parameters:	int a_data_length - The length of the unencoded binary data.
// Returns:		int - The length of the item plus 1 to allow for a terminating
//							\0.
//
// Description: Returns the number of bytes needed to store a base64 encoded
//				array of a_data_length bytes. 
//
//				N.B. The returned length includes space for a terminating \0 
//				so that if a buffer of the returned length is used to store 
//				the encoded version, it can be manipulated as if it were a
//				string using functions such as printf and strcat.

int fumf_enc_get_encoded_byte_array_length( int a_data_length )
{
	return 4 * ( (a_data_length + 2) / 3 ) + 1;	// Allow for terminating \0
}


// Func name:   fumf_enc_generate_encoded_byte_array
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_base64[] - Where the encoded byte Unicode string is to
//							be placed.
//				int a_max_length - The maximum size of the 
//							ac_base64 buffer.
//				const char *ap_data - Pointer to the data to be encoded. 
//							encoded.
//				int a_length - The length of the data in ap_data to be encoded.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Convert binary data stored in a byte array into base64 encoded
//				data suitable for outputing to an encoded message. 
//
//				N.B. This function does NOT pre-pend the ^ character to mark
//				this as binary data.  The user function must do this.
//
//				N.B. If successful, a \0 is appended to the end of the string
//				so that the result can be manipulated as a string by 
//				functions such as printf and strcat.

tumf_bool fumf_enc_generate_encoded_byte_array( sumf_parse_ctrl *ap_parse_ctrl,
												char ac_base64[],
												int a_max_length,
												const char *ap_data,
												int a_length )
{
    int l_in = 0;
	int l_out = 0;
	int l_accumulated = 0;
	int l_number_accumulated = 0;

	
	if( fumf_enc_get_encoded_byte_array_length( a_length ) > a_max_length )
		{
		ap_parse_ctrl->error = EUMF_OUTPUT_TOO_SMALL;
		return FALSE;
		}


	for( l_in=0; l_in < a_length; l_in++ )
		{
		l_accumulated <<= 8;
		l_accumulated |= *ap_data++;
		if( ++l_number_accumulated == 3 )
			{
			p_base64_convert_24bits_to_4x6bits( &(ac_base64[l_out]), l_accumulated );

			l_accumulated = l_number_accumulated = 0;
			l_out += 4;
			}
		}


	if( l_number_accumulated == 1 )
		{
		l_accumulated <<= 2 * 8;

		p_base64_convert_24bits_to_4x6bits( &(ac_base64[l_out]), l_accumulated );

		ac_base64[l_out+2] = ac_base64[l_out+3] = '=';
		}

	else if( l_number_accumulated == 2 )	// Can only be 2 or 3
		{
		l_accumulated <<= 8;

		p_base64_convert_24bits_to_4x6bits( &(ac_base64[l_out]), l_accumulated );

		ac_base64[l_out+3] = '=';
		}

    return TRUE;
}


// Func name:   fumf_enc_get_encoded_embedded_length
//------------------------------------------------------------------------------
// Parameters:	const char ac_embedded[] - Array on which to calculate the 
//							encoded length.
//				int a_embedded_length - The length of the data contained in
//							ac_embedded.
// Returns:		int - The length of the item plus 1 to allow for a terminating
//							\0.
//
// Description: Returns the number of bytes needed to store an escaped embedded
//				message. 
//
//				N.B. The returned length includes space for a terminating \0 
//				so that if a buffer of the returned length is used to store 
//				the encoded version, it can be manipulated as if it were a
//				string using functions such as printf and strcat.

int fumf_enc_get_encoded_embedded_length( const char ac_embedded[], 
											int a_embedded_length )
{
	int l_encoded_embedded_length = 0;
	int l_i;

	for( l_i=0; l_i < a_embedded_length; l_i++ )
		if( ac_embedded[l_i] == '\\' ||
				ac_embedded[l_i] == ')' )
			l_encoded_embedded_length += 2;
		else
			l_encoded_embedded_length++;

	l_encoded_embedded_length++;	// Allow for terminating \0

	return l_encoded_embedded_length;
}


// Func name:   fumf_enc_generate_encoded_embedded
//------------------------------------------------------------------------------
// Parameters:	sumf_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_escaped_embedded[] - Where the escaped embedded 
//							information is to be placed.  
//				int a_escaped_embedded_max_length - The length of the 
//							ac_escaped_embedded buffer.
//				const char ac_embedded[] - The embedded data to be encoded.
// Returns:		tumf_bool - TRUE is successful, FALSE otherwise.  Information
//							on why a function failed can be obtained by calling
//							<fumf_get_last_error>.
//
// Description: Converts a message into the escaped format for
//				outputting on the wire.  
//
//				N.B. If successful, a \0 is appended to the end of the string
//				so that the result can be manipulated as a string by 
//				functions such as printf and strcat.

tumf_bool fumf_enc_generate_encoded_embedded( sumf_parse_ctrl *ap_parse_ctrl,
											char ac_escaped_embedded[], 
											int a_escaped_embedded_max_length,
											const char ac_embedded[] )
{
	int l_i;
	int l_escaped_embedded_length = 0;

	a_escaped_embedded_max_length--;		// Allow for terminating \0	

	for( l_i=0; ac_embedded[l_i] != 0; l_i++ )
		{
		if( ac_embedded[l_i] == '\\' || ac_embedded[l_i] == ')' )
			{
			if( l_escaped_embedded_length + 2 > a_escaped_embedded_max_length )
				{
				ap_parse_ctrl->error = EUMF_OUTPUT_TOO_SMALL;
				return FALSE;
				}

			ac_escaped_embedded[l_escaped_embedded_length++] = '\\';
			ac_escaped_embedded[l_escaped_embedded_length++] = ac_embedded[l_i];
			}
		else
			{
			if( l_escaped_embedded_length + 1 > a_escaped_embedded_max_length )
				{
				ap_parse_ctrl->error = EUMF_OUTPUT_TOO_SMALL;
				return FALSE;
				}

			ac_escaped_embedded[l_escaped_embedded_length++] = ac_embedded[l_i];
			}
		}

	ac_escaped_embedded[l_escaped_embedded_length++] = '\0';

	return TRUE;
}



//------------------------------------------------------------------------------
//                 Copyright Tech-Know-Ware Ltd, 2001, 2002.
//------------------------------------------------------------------------------
